// othercell.cpp

#include "scala.hh"
#include "latsym.hh"
#include "pointgroup.hh"
#include "matvec_utils.hh"
#include "Output.hh"

#define PROGRAM_NAME     std::string("OTHERCELL")
#define PROGRAM_VERSION  std::string("1.1")


using namespace cctbx;
using namespace scala;
using phaser_io::LOGFILE;
using phaser_io::LXML;

//--------------------------------------------------------------

#include "CCP4base.hh"
namespace phaser_io {

  class InterpretCommandLineOC : public InputBase, virtual public CCP4base
  {
  public:
    InterpretCommandLineOC(int argc, char* argv[]);
    InterpretCommandLineOC(Preprocessor& CommandLine);
    Token_value parse(std::istringstream&) {return END;}
    void analyse(void) {}

    // Access
    double Tolerance() const {return tolerance;}

    std::string getXMLOUT() const {return XmloutName;}
      
  private:
    double tolerance;
    std::string XmloutName;

    void initialise(Preprocessor& CommandLine);
  };
//--------------------------------------------------------------
  InterpretCommandLineOC::InterpretCommandLineOC(int argc, char* argv[])
  {
    Preprocessor CommandLine(argc,argv,false);
    initialise(CommandLine);
  }


  InterpretCommandLineOC::InterpretCommandLineOC(Preprocessor& CommandLine)
  {
    initialise(CommandLine);
  }

  void InterpretCommandLineOC::initialise(Preprocessor& CommandLine)
  {
    tolerance = 3.0;
    XmloutName = "";

    parseCCP4(CommandLine);
    // command line has been divide into lines each containing
    // either switch (string beginning "-" or pairs of tokens
    std::istringstream InputStream(const_cast<char*>(CommandLine.Echo().c_str()));

  
    while (get_token(InputStream) != END)
      {
	while (curr_tok != ENDLINE)
	  {
	    if (curr_tok == NAME)
	      {
		if (string_value[0] == '-')
		  {
		    // Switch, ie string beginning with '-'	
		    //		    std::cout << "Command line switch found: " 
		    //			      << string_value << "\n";
		    if (keyIs("-TOL"))
		      {
			// tolerance
			if (get_token(InputStream) != END)
			  {
			    tolerance = number_value;
			  }
		      }
		  }
		else
		  {
		    if (keyIs("XMLOUT"))
		      {
			XmloutName = getString(InputStream);
		      }
		  }
	      }
	    skip_line(InputStream);
	  }
	// Add .xml if needed
	AddFileExtension(XmloutName,"xml");
      }
  }
} // phaser_io

//--------------------------------------------------------------

void PGsort(std::vector<CCtbxSym::PointGroup>& scores)
  // Sort score list on pointgroup "order"
{
  std::sort(scores.begin(), scores.end(), PGcompareOrder());
}

int main(int argc, char* argv[])
{
  // Initialise CCP4 command line parser
  //  CCP4::ccp4fyp(argc, argv);   // NO don't!

  // Initialise output object, CCP4 mode
  phaser_io::Output output;
  output.setPackageCCP4();
  phaser_io::InterpretCommandLineOC CL(argc, argv);
  if (CL.getXMLOUT() != "")
    {output.setXmlout(CL.getXMLOUT());}  // optional XML output
  // Write XML stuff if activated
  time_t start_time;
  std::time(&start_time);
  std::string time = std::string(std::ctime(&start_time));
  time.resize(time.length()-1); // remove trailing \n
  output.logTab(0, LXML,
		 "<"+PROGRAM_NAME+" version=\""+PROGRAM_VERSION+"\"RunTime=\""+
		 time+"\">");

  // Tolerance on cell deviation
  double diff_tolerance = 10.0;
  float maxdel = 10.0;

  try {
    bool eof = false;
    bool target = true;

    std::vector<double> cellin(6);
    std::vector<double> cell_target(6);
    std::vector<double> cell(6);
    output.logTab(0, LOGFILE,"Type cell >>");
    for (int i=0;i<6;i++) 
      {
	std::cin >> cellin[i];
      }

    char LatticeType;
    std::string latstr;
    output.logTab(0, LOGFILE,"Type lattice type >>");
    std::cin >> latstr;
    LatticeType = latstr[0];

    output.logTab(0, LOGFILE,"Type target cell, or eof (control/D) for no target >>");
    for (int i=0;i<6;i++) 
      {
	if (!(std::cin >> cell_target[i]))
	  {
	    eof = true;
	    break;
	  }
      }
    if (eof) target = false;

    output.logTab(0, LOGFILE,"\n\n<<<<<<<<<<<\n\nInput  cell: ");
    output.logTab(0, LOGFILE,FormatCell(cellin,7,4));
    output.logTabPrintf(0, LOGFILE," - Lattice Type %c\n", LatticeType);
    if (target)
      {
	output.logTab(0, LOGFILE,"Target cell:"+FormatCell(cell_target)+"\n");
      }
    // LXML
 
    output.logTab(0, LXML,"<InputCell>\n"+Scell(cellin).xml()+"</InputCell>");
    if (target)
      {output.logTab(0, LXML,"<TargetCell>\n"+Scell(cell_target).xml()+
		      "</TargetCell>\n");}

    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
    // given cellin, LatticeType, construct LatticeSymmetry
    double max_delta = CL.Tolerance();
    int AllowI2 = +1;
    CCtbxSym::LatticeSymmetry lat(cellin, LatticeType, AllowI2, max_delta);

    output.logTab(0, LOGFILE,"Lattice point group: "+
		   lat.best_spacegroup_symbol());
    output.logTabPrintf(0, LOGFILE,
			 " within angular tolerance %6.1f (reset with eg -tol[erance] 5)\n",
			 max_delta);

    // cell permuted & constrained to lattice symmetry      
    cell = lat.best_unit_cell();
    output.logTab(0, LOGFILE,"\n                      Lattice cell: "
	     +FormatCell(cell)+"\n");
    
    // original cell with changed basis but not constrained
    std::vector<double> orig_cell = lat.orig_unit_cell();
    output.logTab(0, LOGFILE,"Lattice unit cell after reindexing: "
	     +FormatCell(orig_cell)
	     +"\n\n");

    // Reindexing operator for "best" spacegroup
    //  from original -> lattice (best), inverse operator for hkl
    ReindexOp reindex_op = MVutil::SetCMat33(lat.best_sg_reindex_op());
    output.logTab(0, LOGFILE,"Reindex operator from initial to lattice cell: "
	     +reindex_op.as_hkl()+"\n\n");
    output.logTab(0, LOGFILE,reindex_op.as_matrix()+"\n");


    output.logTab(0, LXML,"\n<LatticeSymmetry>\n");
    output.logTab(1, LXML,
		   "<LatticegroupName>"+lat.best_spacegroup_symbol()
		   +"</LatticegroupName>");
    output.logTab(1, LXML,Scell(orig_cell).xml());
    output.logTab(0, LXML,"</LatticeSymmetry>\n");
    
    // . . . . . . . . . . . . . . . . . . . . . . . . . . . . 
    // scala::hkl_symmetry object for LatticeSymmetry
    // get list of symmetry elements
    scala::hkl_symmetry symm(lat.best_spacegroup_symbol());
      
    // Make list of subgroups
    std::vector<CCtbxSym::PointGroup> PntGroups = scala::GetSubGroups(symm);
    output.logTab(0, LOGFILE,"\nList of possible Laue groups\n\n");
    output.logTab(0, LOGFILE,"Cell is the cell reindexed into this Laue group\n");
    output.logTab(0, LOGFILE," in a 'standard' setting, unconstrained to group\n");
    output.logTab(0, LOGFILE,"\n'Deviation' is maximum angular deviation (degrees)\n");
    output.logTab(0, LOGFILE," from cell with Laue group symmetry imposed\n");
    if (target) output.logTab(0, LOGFILE,"  or difference from target cell\n");
    output.logTab(0, LOGFILE,"\n'Same cell' and 'Other cell' list alternative indexings\n");
    output.logTab(0, LOGFILE," 'Other cell' lists the cell difference defined as mean square\n");
    output.logTab(0, LOGFILE,"   difference in basis vectors (columns of orthogonalisation matrix)\n");
    if (target)
      output.logTab(0, LOGFILE,"Alternative cells listed are close to target cell, with cell difference\n");
    
    output.logTab(0, LOGFILE,
		   "\nAlternative indexing schemes which lead to identical or similar\n");
    output.logTab(0, LOGFILE," cells are grouped on continuation lines if they are symmetry-related\n\n");
    output.logTab(0, LOGFILE,
		   " Laue group    a     b     c  alpha  beta gamma Deviation ReindexOperator\n");

    output.logTab(0, LXML,"\n\n<LaueGroupList>\n");
    
    scala::ReindexOp RefReindex;
    for (size_t l=0;l<PntGroups.size();l++)
      {
	PntGroups[l].SetCell(cellin, reindex_op, AllowI2);
      }

    // Sort on Laue group "order"  ie spacegroup number
    PGsort(PntGroups);
    std::vector<double> celldiff;
    std::vector<scala::ReindexOp> ReindexList;
    Scell scell(cellin);

    for (size_t l=0;l<PntGroups.size();l++)
      {
	double diff = 0.0;
	if (target)
	  {
	    ReindexList = PntGroups[l].GetCloseCell
	      (scala::Scell(cell_target), diff_tolerance,
	       false, celldiff, AllowI2);
	    if (ReindexList.size() == 0) diff = -1.0;
	  }
	if (diff >= 0.0)
	  {
	    if (target)
	      {
		output.logTabPrintf(0, LXML, "\n<LaueGroupPossibility index=\"%d\">", l);
		// Print all possibilities close to target
		output.logTabPrintf(0, LOGFILE,"\n%10s ",PntGroups[l].RefLGname().c_str());
		for (int i=0;i<6;i++)
		  output.logTabPrintf(0, LOGFILE,"%6.1f",cell_target[i]);
		output.logTabPrintf(0, LOGFILE,"   << Target\n");
		output.logTab(0, LOGFILE,"Alternative cell(s) close to target:\n");
		// XML
		output.logTab(0, LXML,
			       "\n<LaueGroupName>"+PntGroups[l].RefLGname()+"</LaueGroupName>");
		for (size_t k=0; k<ReindexList.size();k++)
		  {
		    output.logTabPrintf(0, LOGFILE,"           ");
		    Scell alt_cell = scell.change_basis(ReindexList[k]);
		    for (int i=0;i<6;i++)
		      output.logTabPrintf(0, LOGFILE,"%6.1f",alt_cell[i]);
		    output.logTabPrintf(0, LOGFILE," %7.2f  ",
					 celldiff.at(k));
		    output.logTab(0, LOGFILE,ReindexList[k].as_hkl()+"\n");
		    // XML
		    output.logTabPrintf(1, LXML,"<ReindexNumber> %4d\n",
k+1);
		    output.logTab(1, LXML,alt_cell.xml());
	
		    output.logTabPrintf(1, LXML,"<CellDelta>%6.2f</CellDelta>\n",
					 celldiff.at(k));
	
		    output.logTab(1, LXML,ReindexList[k].as_hkl_XML()+"\n");
		    output.logTab(1, LXML,ReindexList[k].as_XML());
		    output.logTabPrintf(1, LXML,"</ReindexNumber>\n");
		  }
	      }
	    else
	      // No target cell, print all alternatives
	      {
		output.logTabPrintf(0, LXML, "\n<LaueGroupPossibility index=\"%d\">", l);
		output.logTabPrintf(0, LOGFILE,"\n%10s ",PntGroups[l].RefLGname().c_str());
		std::vector<double> tcell = PntGroups[l].TransformedCell();
		for (int i=0;i<6;i++) output.logTabPrintf(0, LOGFILE,"%6.1f",tcell[i]);
		output.logTabPrintf(0, LOGFILE," %7.2f ",PntGroups[l].Delta());
		output.logTabPrintf(0, LOGFILE," %s",PntGroups[l].RefSGreindexFormat().c_str());
		output.logTabPrintf(0, LOGFILE,"\n");
		// XML
		output.logTab(0, LXML, 
			       "\n<LaueGroupName>"+PntGroups[l].RefLGname()+"</LaueGroupName>");
		output.logTab(1, LXML,Scell(tcell).xml());
		output.logTabPrintf(1, LXML,"<CellDelta>%6.2f</CellDelta>\n",PntGroups[l].
Delta());
	
		output.logTab(1, LXML,PntGroups[l].RefSGreindex().as_hkl_XML()+"\n");
		
		output.logTab(1, LXML,PntGroups[l].RefSGreindex().as_XML());

		PntGroups[l].PrintAlternativeCells(output,true,maxdel,AllowI2);
	      }
	    
	    std::vector<std::string> SGnames =
	      PntGroups[l].SpaceGroupList(Chirality(CHIRAL), true);
	    output.logTab(0, LXML, "<PossibleSpacegroups>\n");
	    for (size_t i = 0; i < SGnames.size(); i++)
	      {
		output.logTab(0, LXML, "<SpaceGroup>" + SGnames[i] +
			       "</SpaceGroup>\n");
	      }
	    output.logTab(0, LXML, "</PossibleSpacegroups>\n");


	    // Is next entry the same lattice group (just check name)?
	    // if so, don't print possible spacegroups yet
	    if (l < PntGroups.size()-1 &&
		PntGroups[l].RefLGname() != PntGroups[l+1].RefLGname())
	      {

		PntGroups[l].PrintAllSpacegroups(output,
					 Chirality(CHIRAL), true);
		output.logTabPrintf(0, LOGFILE,"\n****\n");
	      }
		output.logTabPrintf(0, LXML, "\n</LaueGroupPossibility>",
l);
	  }
      }
    output.logTabPrintf(0, LOGFILE,"\n");
    output.logTab(0, LXML,"</LaueGroupList>\n");


    // - - - - 
  }
  

  catch (std::exception const& err) {
    output.logTab(0, LOGFILE,"\nUNHANDLED EXCEPTION: " +
		   std::string(err.what())+"\n");
  }
  catch (...) {
    output.logTab(0, LOGFILE,"\nUNKNOWN EXCEPTION TYPE\n");
  }

  output.logTab(0, LXML,"</OTHERCELL>\n");



  return 0;
}


